local UH = UnderHood
local UHB = UnderHood_Buffs
local math_min, math_max, math_floor = math.min, math.max, math.floor

local BuffFrame = UHB.FramePrototype
local layouts

--[[
	Smart Layout engine for aura frames.
--]]

function BuffFrame:PerformLayout()
	local buttons = self.buttons
	local layout = layouts[self.config.layout]
	local rowSize = self.config.columns
	local totalButtonsCount, largeButtonsCount, smallButtonsCount = 0, 0, 0
	local firstSmallButtonIndex
	local buttonSize, spacing = self.config.size, self.config.spacing
	
	for i = 1, #buttons do
	    local button = buttons[i]
	    
	    if not buttons[i].id then break end
	    
	    if button.isLarge then
			largeButtonsCount = largeButtonsCount + 1
	    else
	        if not firstSmallButtonIndex then
	            firstSmallButtonIndex = i
			end
			
			smallButtonsCount = smallButtonsCount + 1
		end
	    
	    totalButtonsCount = totalButtonsCount + 1
	end

	if totalButtonsCount == 0 then return end
	
	firstSmallButtonIndex = firstSmallButtonIndex or totalButtonsCount + 1
	local largeButtonIndex = 1
	local smallButtonIndex = firstSmallButtonIndex
	local row, column = 1, 1
	local miniRow, miniColumn, miniLastRowFirstButton, miniRowSize
	local button
	local point, relPoint, dx, dy
	local anchorFrame, lastButton, lastLargeButton, lastSmallButton, lastRowFirstButton
	local buttonsLeft = totalButtonsCount
	local smallButtonsLeft = totalButtonsCount - firstSmallButtonIndex + 1
	local largeButtonsLeft = firstSmallButtonIndex - largeButtonIndex
	
	--while largeButtonIndex < firstSmallButtonIndex or smallButtonIndex < totalButtonsCount+1 do
	while buttonsLeft > 0 do
		local buttonPlaced = false
		
	    if largeButtonIndex < firstSmallButtonIndex and not miniRowSize then
	        -- We have a large button to layout
	        if ((column + 2 < rowSize) or smallButtonsLeft == 0 ) then
	            -- We have a space in the row to place a button
		        button = buttons[largeButtonIndex]
	            button:ClearAllPoints()

		        if column == 1 then
		            anchorFrame = lastRowFirstButton or self.frame
		            point, relPoint, dx, dy = unpack( layout.nextRow )
		            lastRowFirstButton = button
		        else
		            anchorFrame = lastLargeButton
		            point, relPoint, dx, dy = unpack( layout.nextColumn )
		        end

		        button:SetWidth( buttonSize*2 + spacing )
		        button:SetHeight( buttonSize*2 + spacing )
	            button:SetPoint( point, anchorFrame, relPoint , dx*spacing, dy*spacing )

		        lastButton = button
		        lastLargeButton = button

		        column = column + 2
		        largeButtonIndex = largeButtonIndex + 1
		        
		        largeButtonsLeft = largeButtonsLeft - 1

		        if largeButtonsLeft == 0 and column < rowSize then
		            -- There is some space in the row to put small buttons
					miniRow = 1
					miniColumn = 1
					miniRowSize = rowSize - column
					miniLastRowFirstButton = nil
				else
			        buttonPlaced = true
		        end
			else
			    -- We do not have space enough to pace a large button, starting minirow
				miniRow = 1
				miniColumn = 1
				miniRowSize = rowSize - column
				miniLastRowFirstButton = nil
			end
		end
		
		if not buttonPlaced then
			-- No more large buttons or there is a space left in last large row
			button = buttons[smallButtonIndex]
			
			-- Are we in minirow or in the fresh new row?
			if miniRowSize then
			    if miniColumn == 1 then
			        if miniRow == 1 then
			            point, relPoint, dx, dy = unpack( layout.nextColumn )
			            anchorFrame = lastLargeButton
			            miniLastRowFirstButton = button
			        else
			            point, relPoint, dx, dy = unpack( layout.nextRow )
			            anchorFrame = miniLastRowFirstButton
			        end
			    else
			        point, relPoint, dx, dy = unpack( layout.nextColumn )
			        anchorFrame = lastButton
			    end
			    
			    miniColumn = miniColumn + 1
			    
			    if miniColumn > miniRowSize+1 then
			        miniColumn = 1
			        miniRow = miniRow + 1
			        
			        if miniRow > 2 then
			            column = rowSize + 1

			            miniRow = nil
			            miniColumn = nil
			            miniRowSize = nil
			            miniLastRowFirstButton = nil
			        end
				end
			else
				if column == 1 then
				    point, relPoint, dx, dy = unpack( layout.nextRow )
				    anchorFrame = lastRowFirstButton or self.frame
				    lastRowFirstButton = button
				else
				    point, relPoint, dx, dy = unpack( layout.nextColumn )
				    anchorFrame = lastButton
				end
				
				column = column + 1
			end
			
			button:SetWidth( buttonSize )
			button:SetHeight( buttonSize )
			button:ClearAllPoints()
			button:SetPoint( point, anchorFrame, relPoint, dx*spacing, dy*spacing )
			
			lastButton = button
			
			smallButtonIndex = smallButtonIndex + 1
			smallButtonsLeft = smallButtonsLeft - 1
	    end
	    
	    if column > rowSize then
	        column = 1
	        row = row + 1
		end
		
		buttonsLeft = buttonsLeft - 1
	end
end

layouts = {
	RTD = {
	    nextColumn = { "TOPLEFT", "TOPRIGHT", 1, 0 },
	    nextRow = { "TOPLEFT", "BOTTOMLEFT", 0, -1 },
	},
	RTU = {
	    nextColumn = { "TOPLEFT", "TOPRIGHT", 1, 0 },
	    nextRow = { "BOTTOMLEFT", "TOPLEFT", 0, 1 },
	},
	LTD = {
	    nextColumn = { "TOPRIGHT", "TOPLEFT", -1, 0 },
	    nextRow = { "TOPRIGHT", "BOTTOMRIGHT", 0, -1 },
	},
	LTU = {
	    nextColumn = { "TOPRIGHT", "TOPLEFT", -1, 0 },
	    nextRow = { "BOTTOMRIGHT", "TOPRIGHT", 0, 1 },
	},
	DTL = {
	    nextColumn = { "TOPRIGHT", "BOTTOMRIGHT", 0, -1 },
	    nextRow = { "TOPRIGHT", "TOPLEFT", -1, 0 },
	},
	DTR = {
	    nextColumn = { "TOPLEFT", "BOTTOMLEFT", 0, -1 },
	    nextRow = { "TOPLEFT", "TOPRIGHT", 1, 0 },
	},
	UTL = {
	    nextColumn = { "BOTTOMRIGHT", "TOPRIGHT", 0, 1 },
	    nextRow = { "BOTTOMRIGHT", "BOTTOMLEFT", -1, 0 },
	},
	UTR = {
	    nextColumn = { "BOTTOMLEFT", "TOPLEFT", 0, 1 },
	    nextRow = { "BOTTOMLEFT", "BOTTOMRIGHT", 1, 0 },
	},
}
